/*********************************************************************/
// Powers of a square interval matrix A.
//
// Computation of A^k, k= 1,2,... using two different approaches:
//  a) using naive interval computations
//  b) using QR factorizations i.e. using an adapted 
//     coordinate system for each individual step
//     to fight against the wrapping effect.
//
// File: matwrap.cpp
/*********************************************************************/

#include <iostream>
using namespace std;

#include <mvi_util.hpp> //interval matrices/vectors
#include <imath.hpp>    //elementary interval functions
using namespace cxsc;

const int InvFailed= 1;
const int NoError= 0;
const int n= 2;

bool in(const imatrix& A, const imatrix &B) 
//checks whether every component of matrix A is
//contained in the corresponding component of B
{
   bool r(true);
   for (int i=Lb(A,1); i<=Ub(A,1); i++)
     for (int j=Lb(A,2); j<=Ub(A,2); j++)
        r&= in(A[i][j], B[i][j]);
  return r;
}  	

real maxNorm(const rvector& x) //real vector
{
  dotprecision accu(0);
  for (int i=Lb(x); i<=Ub(x); i++)
    accu+= abs(x[i]);		
  return Sup(interval(accu)); //return upper bound
}

real maxNorm(const rmatrix& a) //real matrix
{
  real s=0;
  for (int i=Lb(a,2); i<=Ub(a,2); i++)
    s= Max(s, maxNorm(a[i])); 
  return s;
}

real maxNorm(const ivector& x) //interval vector
{
  dotprecision accu(0);
  for (int i=Lb(x); i<=Ub(x); i++)
    accu+= Sup(abs(x[i]));
  return Sup(interval(accu)); //return upper bound
}

real maxNorm(const imatrix& a)  //interval matrix
{
  real s=0;
  for(int i=Lb(a,2);i<=Ub(a,2);i++)
    s= Max( s, maxNorm(a[i])); 
  return s;
}

rmatrix QfromQR(const rmatrix& A)
{  //QR factorization for 2-by-2 real matrices
  rmatrix Q(2,2);
  Q[1][1]= A[1][1];  Q[1][2]=  A[2][1];
  Q[2][1]= A[2][1];  Q[2][2]= -A[1][1];
  Q/= sqrt( power(A[1][1],2) + power(A[2][1],2) );
  if( A[1][1] > 0.0) Q= -Q;
  return Q;
}

void nearOrthoInv(const rmatrix& Q, imatrix& QInv)
//Compute inverse QInv of a nearly orthogonal matrix Q
{
  int ErrCode= NoError;
  real norm;
  interval det;
  rmatrix Qt( Lb(Q,1),Ub(Q,1),Lb(Q,2),Ub(Q,2) );
    
  ErrCode= NoError;
  Qt= transp(Q);
  norm= maxNorm( Id(Q)-Qt*imatrix(Q) );  
  if (norm < 1.0)
  {
    norm= Sup( maxNorm(Qt)*interval(norm)/(interval(1) - norm) ); 
    det= interval(-norm, norm);
    for (int i=Lb(Q,1); i<=Ub(Q,1); i++)
      for (int j=Lb(Q,2); j<=Ub(Q,2); j++)
        QInv[i][j]= Qt[i][j] + det;
  }
  else
  {
    ErrCode= InvFailed;
    cout << "*** inversion failed! ***" << endl;
    exit(1);
    //QInv= Qt;
  }
}

int main()
{
  cout << boolalpha;
  imatrix A(n,n);
  imatrix Anp1;        //naive iteration
  imatrix QRnp1(n,n);  //improved version using QR
  imatrix Bnp1Inv(n,n);
  rmatrix Bn(n,n), Bnp1(n,n);
  imatrix Yn(n,n), Ynp1(n,n);
  interval phi; 
  
  //numerical example for n=2: rotation matrix
  phi= atan(interval(1)); //enclosure of pi/4
  A[1][1]= cos(phi);   A[1][2]= sin(phi);
  A[2][1]=-sin(phi);   A[2][2]= cos(phi);
    
  cout << "Matrix A: " << endl;
  cout << A << endl;

  //prepare iteration:
  Anp1= A;
  Yn= A;
  Bn= Id(mid(A)); //identity matrix of appropriate dimension
  for (int n=1; n<=125; n++)
  {
    Anp1= A*Anp1; //n==1, then Anp1 == A^2
    
    Bnp1= QfromQR( mid(A*Bn) );
    nearOrthoInv( Bnp1, Bnp1Inv );
    Ynp1= (Bnp1Inv*A*Bn)*Yn;
    QRnp1= Bnp1*Ynp1; //contains A^(n+1)
    //prepare next step
    Bn= Bnp1; 
    Yn= Ynp1;  
    if (n%25 == 0)
    {
      cout << "A^" << n+1 << ":" << endl;
      cout << "Result using naive interval computations:"; 
      cout << endl << SetPrecision(0,9); 
      cout << Anp1;  
      cout << "Improved result using QR factorizations:" << endl;  
      cout << QRnp1;  
    }   
  }
  return 0;
}

/* OUTPUT:

Matrix A:
[  0.707106,  0.707107] [  0.707106,  0.707107]
[ -0.707107, -0.707106] [  0.707106,  0.707107]

A^26:
Result using naive interval computations:
[-0.000000001,0.000000001] [0.999999999,1.000000001]
[-1.000000001,-0.999999999] [-0.000000001,0.000000001]
Improved result using QR factorizations:
[-0.000000001,0.000000001] [0.999999999,1.000000001]
[-1.000000001,-0.999999999] [-0.000000001,0.000000001]
A^51:
Result using naive interval computations:
[-0.707107031,-0.707106531] [0.707106531,0.707107031]
[-0.707107031,-0.707106531] [-0.707107031,-0.707106531]
Improved result using QR factorizations:
[-0.707106782,-0.707106781] [0.707106781,0.707106782]
[-0.707106782,-0.707106781] [-0.707106782,-0.707106781]
A^76:
Result using naive interval computations:
[-1.001445763,-0.998554237] [-0.001445763,0.001445763]
[-0.001445763,0.001445763] [-1.001445763,-0.998554237]
Improved result using QR factorizations:
[-1.000000001,-0.999999999] [-0.000000001,0.000000001]
[-0.000000001,0.000000001] [-1.000000001,-0.999999999]
A^101:
Result using naive interval computations:
[-9.081858848,7.667645285] [-9.081858848,7.667645285]
[-7.667645285,9.081858848] [-9.081858848,7.667645285]
Improved result using QR factorizations:
[-0.707106782,-0.707106781] [-0.707106782,-0.707106781]
[0.707106781,0.707106782] [-0.707106782,-0.707106781]
A^126:
Result using naive interval computations:
[-48511.745857146,48511.745857146] [-48512.745857146,48510.745857146]
[-48510.745857146,48512.745857146] [-48511.745857146,48511.745857146]
Improved result using QR factorizations:
[-0.000000001,0.000000001] [-1.000000001,-0.999999999]
[0.999999999,1.000000001] [-0.000000001,0.000000001]
*/
